from __future__ import annotations

import inspect
import io
import logging
import os
import threading
import uuid
from collections.abc import Iterable
from enum import Enum
from typing import TYPE_CHECKING

from gams.core.gmd import gmdCloseLicenseSession

import gamspy as gp
import gamspy._algebra.expression as expression
import gamspy._algebra.operation as operation
import gamspy._miro as miro
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._backend.backend import backend_factory
from gamspy._convert import GamsConverter, LatexConverter
from gamspy._model_instance import ModelInstance
from gamspy._options import (
    EXECUTION_OPTIONS,
    MODEL_ATTR_OPTION_MAP,
    ModelInstanceOptions,
    Options,
)
from gamspy.exceptions import GamspyException, ValidationError

if TYPE_CHECKING:
    from typing import Literal

    import pandas as pd

    from gamspy import Container, Equation, Parameter, Variable
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation
    from gamspy._backend.engine import EngineClient
    from gamspy._backend.neos import NeosClient
    from gamspy._symbols.implicits import ImplicitParameter, ImplicitVariable
    from gamspy._symbols.symbol import Symbol

IS_MIRO_INIT = os.getenv("MIRO", False)

logger = logging.getLogger("MODEL")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


class Problem(Enum):
    """An enumeration for problem all problem types"""

    LP = "LP"
    NLP = "NLP"
    QCP = "QCP"
    DNLP = "DNLP"
    MIP = "MIP"
    RMIP = "RMIP"
    MINLP = "MINLP"
    RMINLP = "RMINLP"
    MIQCP = "MIQCP"
    RMIQCP = "RMIQCP"
    MCP = "MCP"
    CNS = "CNS"
    MPEC = "MPEC"
    RMPEC = "RMPEC"
    EMP = "EMP"
    MPSGE = "MPSGE"

    @classmethod
    def values(cls):
        """Convenience function to return all values of enum"""
        return list(cls._value2member_map_.keys())

    def __str__(self) -> str:
        return self.value


class Sense(Enum):
    """An enumeration for sense types"""

    MIN = "MIN"
    MAX = "MAX"
    FEASIBILITY = "FEASIBILITY"

    @classmethod
    def values(cls):
        """Convenience function to return all values of enum"""
        return list(cls._value2member_map_.keys())

    def __str__(self) -> str:
        return self.value


class ModelStatus(Enum):
    """An enumeration for model status types"""

    OptimalGlobal = 1
    OptimalLocal = 2
    Unbounded = 3
    InfeasibleGlobal = 4
    InfeasibleLocal = 5
    InfeasibleIntermed = 6
    Feasible = 7
    Integer = 8
    NonIntegerIntermed = 9
    IntegerInfeasible = 10
    LicenseError = 11
    ErrorUnknown = 12
    ErrorNoSolution = 13
    NoSolutionReturned = 14
    SolvedUnique = 15
    Solved = 16
    SolvedSingular = 17
    UnboundedNoSolution = 18
    InfeasibleNoSolution = 19


class SolveStatus(Enum):
    """An enumeration for solve status types"""

    NormalCompletion = 1
    IterationInterrupt = 2
    ResourceInterrupt = 3
    TerminatedBySolver = 4
    EvaluationInterrupt = 5
    CapabilityError = 6
    LicenseError = 7
    UserInterrupt = 8
    SetupError = 9
    SolverError = 10
    InternalError = 11
    Skipped = 12
    SystemError = 13


INTERRUPT_STATUS = [
    SolveStatus.IterationInterrupt,
    SolveStatus.ResourceInterrupt,
    SolveStatus.EvaluationInterrupt,
    SolveStatus.UserInterrupt,
    SolveStatus.TerminatedBySolver,
]

ERROR_STATUS = {
    SolveStatus.CapabilityError: "The solver does not have the capability required by the model.",
    SolveStatus.LicenseError: "The solver cannot find the appropriate license key needed to use a specific subsolver.",
    SolveStatus.SetupError: "The solver encountered a fatal failure during problem set-up time.",
    SolveStatus.SolverError: "The solver encountered a fatal error.",
    SolveStatus.InternalError: "The solver encountered an internal fatal error.",
    SolveStatus.SystemError: "This indicates a completely unknown or unexpected error condition.",
}


# GAMS name -> GAMSPy name
ATTRIBUTE_MAP = {
    "domUsd": "_num_domain_violations",
    "etAlg": "_algorithm_time",
    "etSolve": "_total_solve_time",
    "etSolver": "_total_solver_time",
    "iterUsd": "_num_iterations",
    "marginals": "_marginals",
    "maxInfes": "_max_infeasibility",
    "meanInfes": "_mean_infeasibility",
    "modelStat": "_status",
    "nodUsd": "_num_nodes_used",
    "number": "_solve_number",
    "numDepnd": "_num_dependencies",
    "numDVar": "_num_discrete_variables",
    "numEqu": "_num_equations",
    "numInfes": "_num_infeasibilities",
    "numNLIns": "_num_nonlinear_insts",
    "numNLNZ": "_num_nonlinear_zeros",
    "numNOpt": "_num_nonoptimalities",
    "numNZ": "_num_nonzeros",
    "numRedef": "_num_mcp_redefinitions",
    "numVar": "_num_variables",
    "numVarProj": "_num_bound_projections",
    "objEst": "_objective_estimation",
    "objVal": "_objective_value",
    "procUsed": "_used_model_type",
    "resGen": "_model_generation_time",
    "resUsd": "_solve_model_time",
    "solveStat": "_solve_status",
    "sumInfes": "_sum_infeasibilities",
    "sysVer": "_solver_version",
}


class Model:
    """
    Represents a list of equations to be solved.

    Parameters
    ----------
    container : Container
        Container of the model.
    name : str, optional
        Name of the model. Name is autogenerated by default.
    equations : Iterable[Equation]
        Iterable of Equation objects.
    problem : Problem | str, optional
        'LP', 'NLP', 'QCP', 'DNLP', 'MIP', 'RMIP', 'MINLP', 'RMINLP', 'MIQCP', 'RMIQCP', 'MCP', 'CNS', 'MPEC', 'RMPEC', 'EMP', or 'MPSGE',
        by default Problem.LP.
    sense : Sense | str, optional
        "MIN", "MAX", or "FEASIBILITY". By default, Sense.FEASIBILITY
    objective : Variable | Expression, optional
        Objective variable to minimize or maximize or objective itself.
    matches : dict[Equation, Variable]
        Equation - Variable matches for MCP models.
    limited_variables : Iterable[ImplicitVariable], optional
        Allows limiting the domain of variables used in a model.
    external_module: str, optional
        The name of the external module in which the external equations are implemented

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", definition= v == 5)
    >>> my_model = gp.Model(m, "my_model", "LP", [e])

    """

    # Prefix for auto-generated symbols
    _generate_prefix = "autogenerated_"
    _autogen_symbols: list[Variable | Equation] = []

    def __init__(
        self,
        container: Container | None = None,
        name: str | None = None,
        problem: Problem | str = Problem.MIP,
        equations: Iterable[Equation] = [],
        sense: Sense | str = Sense.FEASIBILITY,
        objective: Variable | Expression | None = None,
        matches: dict[Equation, Variable] | None = None,
        limited_variables: Iterable[ImplicitVariable] | None = None,
        external_module: str | None = None,
    ):
        self._auto_id = "m" + str(uuid.uuid4()).replace("-", "_")

        if name is not None:
            name = validation.validate_name(name)
            self.name = validation.validate_model_name(name)
        else:
            self.name = self._auto_id

        if container is not None:
            self.container = container
        else:
            self.container = gp._ctx_managers[
                (os.getpid(), threading.get_native_id())
            ]

        assert self.container is not None
        self._matches = matches
        self.problem, self.sense = validation.validate_model(
            equations, problem, sense
        )
        self.equations = list(equations)
        self._objective = objective
        self._objective_variable = self._set_objective_variable(objective)
        if (
            self._objective_variable is not None
            and self._objective_variable.type.lower() != "free"
        ):
            raise ValidationError(
                f"Objective variable `{self._objective_variable}` must be a free variable"
            )

        if limited_variables is not None and not isinstance(
            limited_variables, Iterable
        ):
            raise ValidationError(
                f"`limited_variables must be an Iterable of ImplicitVariable objects but found {limited_variables}`"
            )

        self._limited_variables = limited_variables

        if not self.equations and not self._matches:
            raise ValidationError("Model requires at least one equation.")

        self._external_module_file: str | None = None
        self._external_module: str | None = None

        if external_module is not None:
            self.external_module = external_module
        else:
            # To avoid adding it twice
            self.container._add_statement(self)

        # allow freezing
        self._is_frozen: bool = False

        # Attributes
        self._num_domain_violations = None
        self._algorithm_time = None
        self._total_solve_time = None
        self._total_solver_time = None
        self._num_iterations = None
        self._marginals = None
        self._max_infeasibility = None
        self._mean_infeasibility = None
        self._status: ModelStatus | None = None
        self._num_nodes_used = None
        self._solve_number = None
        self._num_dependencies = None
        self._num_discrete_variables = None
        self._num_equations = None
        self._num_infeasibilities = None
        self._num_nonlinear_insts = None
        self._num_nonlinear_zeros = None
        self._num_nonoptimalities = None
        self._num_nonzeros = None
        self._num_mcp_redefinitions = None
        self._num_variables = None
        self._num_bound_projections = None
        self._objective_estimation = None
        self._objective_value = None
        self._used_model_type = None
        self._model_generation_time = None
        self._solve_model_time = None
        self._sum_infeasibilities = None
        self._solve_status: SolveStatus | None = None
        self._solver_version = None

        self.container._synch_with_gams()

    def __repr__(self) -> str:
        return f"Model(name='{self.name}', problem='{str(self.problem)}', equations={self.equations}, sense='{str(self.sense)}', objective={self._objective_variable}, matches={self._matches}, limited_variables={self._limited_variables}"

    def __str__(self) -> str:
        return (
            f"Model {self.name}:\n  Problem Type: {self.problem}\n  Sense:"
            f" {self.sense}\n  Equations: {self.equations}"
        )

    @property
    def num_domain_violations(self) -> float | None:
        """
        Number of domain violations.

        Returns
        -------
        float | None
        """
        return self._num_domain_violations

    @property
    def algorithm_time(self) -> float | None:
        """
        Solver dependent timing information. This attribute was intended to allow
        solvers to return the elapsed time used by the solve algorithm without
        including any model generation, communication, or setup time. However,
        solvers are free to adapt this convention and return time-related information
        (but not necessarily elapsed time) for executing the solve algorithm. Please
        inspect your solver manual for the actual meaning of the value returned in
        this attribute.

        Returns
        -------
        float | None
        """
        return self._algorithm_time

    @property
    def total_solve_time(self) -> float | None:
        """
        Elapsed time it took to execute a solve statement in total. This model
        attribute returns the elapsed time it took to execute a solve statement in total.
        This time includes the model generation time, the time to read and write files,
        the time to create the solution report and the time taken by the actual solve.
        The time is expressed in seconds of wall-clock time.

        Returns
        -------
        float | None
        """
        return self._total_solve_time

    @property
    def total_solver_time(self) -> float | None:
        """
        Elapsed time taken by the solver only. This model attribute returns the elapsed
        time taken by the solver only. This does not include the GAMS model generation time
        and the time taken to report and load the solution back into the GAMS database.
        The time is expressed in seconds of wall-clock time.

        Returns
        -------
        float | None
        """
        return self._total_solver_time

    @property
    def num_iterations(self) -> float | None:
        """
        Number of iterations used.

        Returns
        -------
        float | None
        """
        return self._num_iterations

    @property
    def marginals(self) -> float | None:
        """
        Indicates whether there are marginals.

        Returns
        -------
        float | None
        """
        return self._marginals

    @property
    def max_infeasibility(self) -> float | None:
        """
        Maximum of infeasibilities

        Returns
        -------
        float | None
        """
        return self._max_infeasibility

    @property
    def mean_infeasibility(self) -> float | None:
        """
        Mean of infeasibilities

        Returns
        -------
        float | None
        """
        return self._mean_infeasibility

    @property
    def status(self) -> ModelStatus | None:
        """
        Model status after solve.

        Returns
        -------
        ModelStatus | None
        """
        return self._status

    @property
    def num_nodes_used(self) -> float | None:
        """
        Number of nodes used by the MIP solver.

        Returns
        -------
        float | None
        """
        return self._num_nodes_used

    @property
    def solve_number(self) -> float | None:
        """
        Number of the last solve.

        Returns
        -------
        float | None
        """
        return self._solve_number

    @property
    def num_dependencies(self) -> float | None:
        """
        Number of dependencies in a CNS model.

        Returns
        -------
        float | None
        """
        return self._num_dependencies

    @property
    def num_discrete_variables(self) -> float | None:
        """
        Number of discrete variables.

        Returns
        -------
        float | None
        """
        return self._num_discrete_variables

    @property
    def num_equations(self) -> float | None:
        """
        Number of equations.

        Returns
        -------
        float | None
        """
        return self._num_equations

    @property
    def num_infeasibilities(self) -> float | None:
        """
        Number of infeasibilities.

        Returns
        -------
        float | None
        """
        return self._num_infeasibilities

    @property
    def num_nonlinear_insts(self) -> float | None:
        """
        Number of nonlinear instructions.

        Returns
        -------
        float | None
        """
        return self._num_nonlinear_insts

    @property
    def num_nonlinear_zeros(self) -> float | None:
        """
        Number of nonlinear nonzeros.

        Returns
        -------
        float | None
        """
        return self._num_nonlinear_zeros

    @property
    def num_nonoptimalities(self) -> float | None:
        """
        Number of nonoptimalities.

        Returns
        -------
        float | None
        """
        return self._num_nonoptimalities

    @property
    def num_nonzeros(self) -> float | None:
        """
        Number of nonzero entries in the model coefficient matrix.

        Returns
        -------
        float | None
        """
        return self._num_nonzeros

    @property
    def num_mcp_redefinitions(self) -> float | None:
        """
        Number of MCP redefinitions.

        Returns
        -------
        float | None
        """
        return self._num_mcp_redefinitions

    @property
    def num_variables(self) -> float | None:
        """
        Number of variables.

        Returns
        -------
        float | None
        """
        return self._num_variables

    @property
    def num_bound_projections(self) -> float | None:
        """
        Number of bound projections during model generation.

        Returns
        -------
        float | None
        """
        return self._num_bound_projections

    @property
    def objective_estimation(self) -> float | None:
        """
        Estimate of the best possible solution for a mixed-integer model

        Returns
        -------
        float | None
        """
        return self._objective_estimation

    @property
    def objective_value(self) -> float | None:
        """
        Objective function value

        Returns
        -------
        float | None
        """
        return self._objective_value

    @property
    def used_model_type(self) -> float | None:
        """
        Integer number that indicates the used model type.

        Returns
        -------
        float | None
        """
        return self._used_model_type

    @property
    def model_generation_time(self) -> float | None:
        """
        Time GAMS took to generate the model in wall-clock seconds.

        Returns
        -------
        float | None
        """
        return self._model_generation_time

    @property
    def solve_model_time(self) -> float | None:
        """
        Time the solver used to solve the model in seconds

        Returns
        -------
        float | None
        """
        return self._solve_model_time

    @property
    def solve_status(self) -> SolveStatus | None:
        """
        Indicates the solver termination condition.

        Returns
        -------
        SolveStatus | None
        """
        return self._solve_status

    @property
    def sum_infeasibilities(self) -> float | None:
        """
        Sum of infeasibilities.

        Returns
        -------
        float | None
        """
        return self._sum_infeasibilities

    @property
    def solver_version(self) -> float | None:
        """
        Solver version.

        Returns
        -------
        float | None
        """
        return self._solver_version

    @property
    def external_module(self) -> str | None:
        """
        Name of the external module in which the external equations are implemented.
        By default, this parameter is set to None. When provided, it triggers the
        opening of the specified file using a File statement and incorporates the
        file into the model by adding it as an external module.

        This feature requires a solid understanding of programming, compilation,
        and linking processes. For more information, please refer to the
        https://gamspy.readthedocs.io/en/latest/user/advanced/external_equations.html .

        Returns
        -------
        str | None
        """
        return self._external_module

    @external_module.setter
    def external_module(self, value: str | None):
        if self._external_module_file is not None:
            self.container._add_statement(
                f"putclose {self._external_module_file};"
            )

        write_model_statement = value != self._external_module
        self._external_module = None
        self._external_module_file = None

        if value is not None:
            filename = "f" + str(uuid.uuid4()).replace("-", "_")
            self._external_module_file = filename
            self._external_module = value
            self.container._add_statement(f"File {filename} / '{value}' /;")

        if write_model_statement:
            self.container._add_statement(self)

    def _generate_obj_var_and_equation(self) -> tuple[Variable, Equation]:
        variable = gp.Variable._constructor_bypass(
            self.container,
            f"{self.name}_objective_variable",
            domain=[],
        )
        equation = gp.Equation._constructor_bypass(
            self.container,
            f"{self.name}_objective",
            domain=[],
        )
        self._autogen_symbols.append(variable.name)
        self._autogen_symbols.append(equation.name)

        return variable, equation

    def _set_objective_variable(
        self,
        assignment: None | Variable | Operation | Expression = None,
    ) -> Variable | None:
        """
        Returns objective variable. If the assignment is an Expression
        or an Operation (Sum, Product etc.), it automatically generates
        an objective variable and a equation.

        Returns
        -------
        Variable | None

        Raises
        ------
        TypeError
            In case assignment is not a Variable, Expression or an Operation.
        """

        if assignment is not None and not isinstance(
            assignment,
            (gp.Variable, expression.Expression, operation.Operation),
        ):
            raise TypeError(
                "Objective must be a Variable or an Expression but"
                f" {type(assignment)} given"
            )

        if self.sense == gp.Sense.FEASIBILITY:
            if assignment is not None:
                raise ValidationError(
                    "Cannot set an objective when the sense is FEASIBILITY!"
                )

            if self.problem in [Problem.CNS, Problem.MCP, Problem.EMP]:
                return None

            variable, equation = self._generate_obj_var_and_equation()
            statement = expression.Expression(
                implicits.ImplicitEquation(
                    equation,
                    name=equation.name,
                    type=equation.type,
                    domain=[],
                ),
                "..",
                variable == 0,
            )
            self.container._add_statement(statement)
            equation._definition = statement
            equation.modified = False
            variable.modified = False

            if equation.name not in [symbol.name for symbol in self.equations]:
                self.equations.append(equation)

            return variable

        if isinstance(
            assignment, (expression.Expression, operation.Operation)
        ):
            variable, equation = self._generate_obj_var_and_equation()

            # Sum((i,j),c[i,j]*x[i,j])->Sum((i,j),c[i,j]*x[i,j]) =e= var
            assignment = assignment == variable

            # equation .. Sum((i,j),c[i,j]*x[i,j]) =e= var
            statement = expression.Expression(
                implicits.ImplicitEquation(
                    equation,
                    name=equation.name,
                    type=equation.type,
                    domain=[],
                ),
                "..",
                assignment,
            )
            self.container._add_statement(statement)
            equation._definition = statement
            equation.modified = False
            variable.modified = False
            if equation.name not in [symbol.name for symbol in self.equations]:
                self.equations.append(equation)

            return variable

        return assignment

    def _generate_solve_string(self) -> str:
        solve_string = f"solve {self.name} using {self.problem}"

        if self.sense == gp.Sense.FEASIBILITY:
            # Set sense as min or max for feasibility
            self.sense = gp.Sense("MIN")

        if self.problem not in [Problem.MCP, Problem.CNS, Problem.EMP]:
            solve_string += f" {self.sense}"

        if self._objective_variable is not None:
            solve_string += f" {self._objective_variable.gamsRepr()}"

        return solve_string + ";"

    def _add_runtime_options(
        self, options: Options, backend: str = "local"
    ) -> None:
        for key, value in options.model_dump(exclude_none=True).items():
            if key in MODEL_ATTR_OPTION_MAP:
                if isinstance(value, bool):
                    value = int(value)
                elif isinstance(value, str):
                    value = f"'{value}'"

                self.container._add_statement(
                    f"{self.name}.{MODEL_ATTR_OPTION_MAP[key]} = {value};\n"
                )
            elif key in EXECUTION_OPTIONS:
                if backend == "engine" and key == "loadpoint":
                    value = os.path.relpath(
                        value, self.container.working_directory
                    )

                self.container._add_statement(
                    f"{EXECUTION_OPTIONS[key]} '{value}';\n"
                )

    def _append_solve_string(self) -> None:
        solve_string = self._generate_solve_string()
        self.container._add_statement(solve_string + "\n")

    def _create_model_attributes(self) -> None:
        self.container._add_statement("$offListing")
        for attr_name in ATTRIBUTE_MAP:
            symbol_name = f"{self._generate_prefix}{attr_name}_{self._auto_id}"
            _ = gp.Parameter._constructor_bypass(self.container, symbol_name)

            self.container._add_statement(
                f"{symbol_name} = {self.name}.{attr_name};"
            )
        self.container._add_statement("$onListing")

    def _update_model_attributes(self) -> None:
        container = self.container._temp_container
        gdx_handle = utils._open_gdx_file(
            self.container.system_directory, self.container._gdx_out
        )

        for gams_attr, python_attr in ATTRIBUTE_MAP.items():
            symbol_name = f"{self._generate_prefix}{gams_attr}_{self._auto_id}"
            data = utils._get_scalar_data(
                container._gams2np, gdx_handle, symbol_name
            )

            if python_attr == "_status":
                setattr(self, python_attr, ModelStatus(data))
            elif python_attr == "_solve_status":
                status = SolveStatus(data)
                setattr(self, python_attr, status)

                if status in INTERRUPT_STATUS:
                    logger.warning(
                        f"The solve was interrupted! Solve status: {status.name}. "
                        "For further information, see https://gamspy.readthedocs.io/en/latest/reference/gamspy._model.html#gamspy.SolveStatus."
                    )
                elif status in ERROR_STATUS:
                    raise GamspyException(
                        f"Solve status: {status.name}. {ERROR_STATUS[status]}",
                        status.value,
                    )
            else:
                setattr(self, python_attr, data)

        utils._close_gdx_handle(gdx_handle)
        self.container._temp_container.data = {}

    def computeInfeasibilities(self) -> dict[str, pd.DataFrame]:
        """
        Computes infeasabilities for all equations of the model

        Returns
        -------
        dict[str, pd.DataFrame]
            Dictionary of infeasibilities where equation names are keys and
            infeasibilities are values

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["i1", "i2"])
        >>> j = gp.Set(m, name="j", records=["j1", "j2", "j3"])
        >>> a = gp.Parameter(m, name="a", domain=i, records=[("i1", 350), ("i2", 600)])
        >>> b = gp.Parameter(m, name="b", domain=j, records=[("j1", 400), ("j2", 450), ("j3", 420)])
        >>> x = gp.Variable(m, name="x", domain=[i,j], type="Positive")
        >>> s = gp.Equation(m, name="s", domain=i)
        >>> d = gp.Equation(m, name="d", domain=j)
        >>> s[i] = gp.Sum(j, x[i, j]) <= a[i]
        >>> d[j] = gp.Sum(i, x[i, j]) >= b[j]
        >>> my_model = gp.Model(m, name="my_model", equations=m.getEquations(), problem="LP", sense="min", objective=gp.Sum((i, j), x[i, j]))
        >>> summary = my_model.solve()
        >>> infeasibilities = my_model.computeInfeasibilities()
        >>> infeasibilities["s"].infeasibility.item()
        320.0

        """
        infeas_dict = {}

        for equation in self.equations:
            if equation.records is None:
                continue

            infeas_rows = utils._calculate_infeasibilities(equation)
            infeas_dict[equation.name] = infeas_rows

            if equation._definition is not None:
                names = equation._definition._find_all_symbols()
                names = [
                    name
                    for name in names
                    if name in self.container.data
                    and isinstance(self.container[name], gp.Variable)
                ]

                for name in names:
                    variable = self.container[name]
                    infeas_rows = utils._calculate_infeasibilities(variable)
                    infeas_dict[variable.name] = infeas_rows

        return infeas_dict

    def getEquationListing(
        self,
        infeasibility_threshold: float | None = None,
    ) -> str:
        """
        Returns the generated equations.

        Parameters
        ----------
        infeasibility_threshold: float, optional
            Filters out equations with infeasibilities that are above this value.

        Returns
        -------
        str
        """
        listings = []
        for equation in self.equations:
            listing = equation.getEquationListing(
                infeasibility_threshold=infeasibility_threshold
            )

            if listing:
                listings.append(listing)

        return "\n".join(listings)

    def getVariableListing(self) -> str:
        """
        Returns the variable listing.

        Returns
        -------
        str
        """
        if not hasattr(self, "_variables"):
            raise ValidationError(
                "The model must be solved with `variable_listing_limit` option for this functionality to work."
            )

        listings = []
        for variable in self._variables:
            listings.append(variable.getVariableListing())

        return "\n".join(listings)

    def interrupt(self) -> None:
        """Sends interrupt signal to the running job."""
        self.container._interrupt()

    def freeze(
        self,
        modifiables: list[Parameter | ImplicitParameter],
        options: Options | None = None,
    ) -> None:
        self._is_frozen = True
        if options is None:
            options = Options()

        self.instance = ModelInstance(
            self.container, self, modifiables, options
        )

    def unfreeze(self) -> None:
        """Unfreezes the model"""
        self._is_frozen = False
        gmdCloseLicenseSession(self.instance.sync_db.gmd)

    def solve(
        self,
        solver: str | None = None,
        options: Options | None = None,
        solver_options: dict | None = None,
        model_instance_options: ModelInstanceOptions | None = None,
        output: io.TextIOWrapper | None = None,
        backend: Literal["local", "engine", "neos"] = "local",
        client: EngineClient | NeosClient | None = None,
        load_symbols: list[Symbol] | None = None,
    ) -> pd.DataFrame | None:
        """
        Solves the model with given options.

        Parameters
        ----------
        solver : str, optional
            Solver name
        options : Options, optional
            GAMS options
        solver_options : dict, optional
            Solver options
        model_instance_options : ModelInstanceOptions, optional
            Model instance options
        output : TextIOWrapper, optional
            Output redirection target
        backend : str, optional
            Backend to run on
        client : EngineClient, NeosClient, optional
            EngineClient to communicate with GAMS Engine or NEOS Client to communicate with NEOS Server
        load_symbols : list[Symbol], optional
            Specifies the symbols that need to be loaded. If not given, all symbols are loaded after solve.

        Returns
        -------
        DataFrame, optional
            Summary of the solve

        Raises
        ------
        ValidationError
            In case engine_config is not provided for `engine` backend or
            neos_client is not provided for `neos` backend.
        ValueError
            In case problem is not in possible problem types
        ValueError
            In case sense is different than "MIN" or "MAX"

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", "LP", [e], "max", v)
        >>> solved = my_model.solve()

        """
        if solver is None:
            solver = utils.getDefaultSolvers()[str(self.problem).upper()]

        validation.validate_solver_args(
            self.container.system_directory,
            backend,
            solver,
            self.problem,
            options,
            output,
            load_symbols,
        )
        validation.validate_equations(self)

        if options is None:
            options = self.container._options

        # Only for local until GAMS Engine and NEOS Server backends adopt the new GP_SolveLine option.
        if solver == "local":
            frame = inspect.currentframe().f_back
            options._frame = frame

        if self._is_frozen:
            if model_instance_options is None:
                model_instance_options = ModelInstanceOptions()

            summary = self.instance.solve(
                solver, model_instance_options, solver_options, output
            )
            return summary

        self.container._add_statement(self.getDeclaration())
        self._add_runtime_options(options, backend)
        self._append_solve_string()
        self._create_model_attributes()
        options._set_solver_options(
            working_directory=self.container.working_directory,
            solver=solver,
            problem=self.problem,
            solver_options=solver_options,
        )

        runner = backend_factory(
            self.container,
            options,
            solver,
            solver_options,
            output,
            backend,
            client,
            self,
            load_symbols,
        )

        summary = runner.run(relaxed_domain_mapping=False, gams_to_gamspy=True)

        if IS_MIRO_INIT:
            miro._write_default_gdx_miro(self.container)

        return summary

    def getDeclaration(self) -> str:
        """
        Declaration of the Model in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", "LP", [e])
        >>> my_model.getDeclaration()
        'Model my_model / e,my_model_objective /;'

        """
        equations = []
        for equation in self.equations:
            if self._matches:
                if equation not in self._matches:
                    equations.append(equation.name)
            else:
                equations.append(equation.name)

        equations_str = ",".join(equations)

        if self._matches:
            matches_str = ",".join(
                [
                    f"{equation.name}.{variable.name}"
                    for equation, variable in self._matches.items()
                ]
            )

            equations_str = (
                ",".join([equations_str, matches_str])
                if equations
                else matches_str
            )

        if self._external_module_file:
            equations_str = ",".join(
                [equations_str, self._external_module_file]
            )

        if self._limited_variables:
            limited_variables_str = ",".join(
                [variable.gamsRepr() for variable in self._limited_variables]
            )
            equations_str += "," + limited_variables_str

        model_str = f"Model {self.name}"
        if equations_str != "":
            model_str += f" / {equations_str} /"
        model_str += ";"

        return model_str

    def toGams(self, path: str, options: Options | None = None) -> None:
        """
        Generates GAMS model under path/<model_name>.gms

        Parameters
        ----------
        path : str
            Path to the directory which will contain the GAMS model.
        """
        if options is not None and not isinstance(options, gp.Options):
            raise ValidationError(
                f"`options` must be of type gp.Options of found {type(options)}"
            )

        converter = GamsConverter(self, path)
        converter.convert(options=options)

    def toLatex(self, path: str, generate_pdf: bool = False) -> None:
        """
        Generates a latex file that contains the model definition under path/<model_name>.tex

        Parameters
        ----------
        path : str
            Path to the directory which will contain the .tex file.
        """
        converter = LatexConverter(self, path)
        converter.convert()

        if generate_pdf:
            converter.to_pdf()



from __future__ import annotations

import logging
import os
import subprocess
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import gamspy._symbols as syms
import gamspy.utils as utils
from gamspy._options import EXECUTION_OPTIONS, MODEL_ATTR_OPTION_MAP, Options
from gamspy.exceptions import LatexException, ValidationError

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from gamspy import Alias, Equation, Model, Parameter, Set, Variable

    SymbolType: TypeAlias = Alias | Set | Parameter | Variable | Equation

logger = logging.getLogger("CONVERTER")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


def get_symbols(model) -> list[str]:
    all_symbols = []
    for equation in model.equations:
        symbols = equation._definition._find_all_symbols()

        for symbol in symbols:
            if symbol not in all_symbols:
                all_symbols.append(symbol)

    if model._matches:
        for equation in model._matches:
            symbols = equation._definition._find_all_symbols()
            for symbol in symbols:
                if symbol not in all_symbols:
                    all_symbols.append(symbol)

    return all_symbols


class Converter(ABC):
    @abstractmethod
    def convert(self): ...


class GamsConverter(Converter):
    def __init__(self, model: Model, path: str) -> None:
        os.makedirs(path, exist_ok=True)
        self.model = model
        self.container = model.container
        self.path = path
        self.gdx_path = os.path.join(path, model.name + "_data.gdx")
        self.gms_path = os.path.join(path, model.name + ".gms")

    def get_definitions(self) -> list[str]:
        definitions = []
        for equation in self.model.equations:
            assert equation._definition is not None
            definitions.append(equation._definition.getDeclaration())

        if self.model._matches:
            for equation in self.model._matches:
                assert equation._definition is not None
                definitions.append(equation._definition.getDeclaration())

        return definitions

    def get_all_symbols(self) -> list[str]:
        all_symbols = get_symbols(self.model)

        all_needed_symbols = sorted(
            all_symbols, key=list(self.container.data.keys()).index
        )

        return all_needed_symbols

    def convert(self, options: Options | None = None) -> None:
        """Generates .gms and .gdx file"""
        symbols = self.get_all_symbols()

        # Write the symbol data first
        self.container.write(self.gdx_path, symbols)

        # 1. Declarations
        declarations = [
            self.container[name].getDeclaration() for name in symbols
        ]

        if self.model.external_module is not None:
            em_name = self.model._external_module
            em_file = self.model._external_module_file
            declarations.append(f"File {em_file} /{em_name}/;")
            logger.info("Converter will not copy external module files")
            logger.info(
                f"You need to ensure your external module is accessible from {self.path}"
            )

        # 2. Load the data from gdx
        load_str = f"$gdxLoadAll {os.path.abspath(self.gdx_path)}"

        # 3. Definitions
        definitions = self.get_definitions()
        declarations.append(self.model.getDeclaration())

        # 4. Model attribute options
        options_strs = []
        if options is not None:
            options._export(os.path.join(self.path, f"{self.model.name}.pf"))
            for key, value in options.model_dump(exclude_none=True).items():
                if key in MODEL_ATTR_OPTION_MAP:
                    if isinstance(value, bool):
                        value = int(value)
                    elif isinstance(value, str):
                        value = f"'{value}'"

                    options_strs.append(
                        f"{self.model.name}.{MODEL_ATTR_OPTION_MAP[key]} = {value};"
                    )
                elif key in EXECUTION_OPTIONS:
                    options_strs.append(f"{EXECUTION_OPTIONS[key]} '{value}'")

        # 5. Solve string
        solve_string = self.model._generate_solve_string()
        strings = [
            *declarations,
            load_str,
            *definitions,
            *options_strs,
            solve_string,
        ]

        # Write the GAMS code
        gams_string = "\n".join(strings)
        with open(self.gms_path, "w", encoding="utf-8") as file:
            file.write(gams_string)

        logger.info(
            f'GAMS (.gms) file has been generated under {os.path.join(self.path, self.model.name + ".gms")}'
        )


TABLE_HEADER = """\\begin{tabularx}{\\textwidth}{| l | l | X |}
\\hline
\\textbf{Name} & \\textbf{Domains} & \\textbf{Description}\\\\
\\hline
\\endhead
"""

TABLE_FOOTER = "\\hline\n\\end{tabularx}"


class LatexConverter(Converter):
    def __init__(self, model: Model, path: str) -> None:
        os.makedirs(path, exist_ok=True)
        self.model = model
        self.container = model.container
        self.path = path
        symbols = get_symbols(model)
        if self.model._limited_variables:
            for var in self.model._limited_variables:
                domain = var.domain
                for elem in domain:
                    if (
                        isinstance(elem, (syms.Set, syms.Alias))
                        and elem.name not in symbols
                    ):
                        symbols.append(elem.name)

        self.symbols = sorted(
            symbols, key=list(self.container.data.keys()).index
        )

        for name in self.model._autogen_symbols:
            if name in self.symbols:
                self.symbols.remove(name)

        self.header = self.get_header()
        self.set_header = "\\subsection*{Sets}"
        self.param_header = "\\subsection*{Parameters}"
        self.variable_header = "\\subsection*{Variables}"
        self.equation_header = "\\subsection*{Equations}"
        self.definitions_header = "\\subsection*{Model Definition}"
        self.footer = "\\end{document}"
        self.tex_path = os.path.join(path, model.name + ".tex")

    def convert(self) -> None:
        latex_strs = [self.header]

        # Sets
        latex_strs.append(self.set_header)
        latex_strs.append(self.get_table((syms.Set, syms.Alias)))

        # Parameters
        latex_strs.append(self.param_header)
        latex_strs.append(self.get_table(syms.Parameter))

        # Variables
        latex_strs.append(self.variable_header)
        latex_strs.append(self.get_table(syms.Variable))

        # Equations
        latex_strs.append(self.equation_header)
        latex_strs.append(self.get_table(syms.Equation))

        # Definitions
        latex_strs.append(self.definitions_header)
        if self.model._objective is None:
            ...
        elif isinstance(self.model._objective, syms.Variable):
            latex_strs.append(
                f"\\textbf{{{str(self.model.sense).lower()}}} ${self.model._objective_variable.name}$\\\\"  # type: ignore
            )
            latex_strs.append("\\textbf{s.t.}")
        else:
            latex_strs.append(
                f"\\textbf{{{str(self.model.sense).lower()}}} ${self.model._objective.latexRepr()}$\\\\"
            )
            latex_strs.append("\\textbf{s.t.}")

        latex_strs.append(self.get_definitions())

        # Constraints
        latex_strs.append(self.get_constraints())

        latex_strs.append(self.footer)

        latex_str = "\n".join(latex_strs)
        with open(
            self.tex_path, "w", encoding="utf-8"
        ) as file:  # Write the TEX file
            file.write(latex_str)

        logger.info(
            f'LaTeX (.tex) file has been generated under {os.path.join(self.path, self.model.name + ".tex")}'
        )

        self.latex_str = latex_str

    def to_pdf(self) -> None:
        process = subprocess.run(["pdflatex", "-v"])
        if process.returncode:
            raise ValidationError(
                "`pdflatex` is required to generate the pdf! Please install `pdflatex` and add it to the path."
            )

        process = subprocess.run(
            ["pdflatex", f"-output-directory={self.path}", self.tex_path],
            capture_output=True,
            text=True,
        )
        if process.returncode:
            raise LatexException(
                f"Could not generate pdf file: {process.stderr}"
            )

    def get_table(self, symbol_type):
        table = [TABLE_HEADER]
        for name in self.symbols:
            symbol: SymbolType = self.container[name]

            if isinstance(symbol, symbol_type):
                summary = symbol.summary
                domain_str = ",".join(summary["domain"])
                if (
                    isinstance(symbol, syms.Variable)
                    and self.model._limited_variables
                ):
                    for elem in self.model._limited_variables:
                        if elem.name == symbol.name:
                            domain_str = utils._get_domain_str(elem.domain)[
                                1:-1
                            ]

                row = f'{summary["name"]} & {domain_str} & {summary["description"]}\\\\'
                row = row.replace("_", "\\_")
                table.append(row)

        table.append(TABLE_FOOTER)

        return "\n".join(table)

    def get_definitions(self) -> str:
        definitions = []
        for equation in self.model.equations:
            if equation.name in self.model._autogen_symbols:
                continue

            domain_str = ",".join([elem.name for elem in equation.domain])
            header = "\\subsubsection*{$" + equation.name.replace("_", "\\_")
            if domain_str:
                header += f"_{{{domain_str}}}"
            header += "$}\n"

            footer = "\n\\vspace{5pt}\n\\hrule"
            latex_repr = f"{header}{equation.latexRepr()}{footer}"
            definitions.append(latex_repr)

        return "\n".join(definitions)

    def get_constraints(self) -> str:
        constraints = ["\\bigskip"]
        for name in self.symbols:
            symbol = self.container[name]
            if not isinstance(symbol, syms.Variable):
                continue

            constraint = "$" + symbol.latexRepr()
            if symbol.type == "binary":
                constraint += (
                    "\\in "
                    + r"\{0,1\}"
                    + " ~ \\forall "
                    + ",".join(symbol.domain_names)
                )
            elif symbol.type == "integer":
                constraint += "\\in \\mathbb{Z}_{+} \\forall " + ",".join(
                    symbol.domain_names
                )
            elif symbol.type == "positive":
                constraint += "\\geq 0 ~ \\forall " + ",".join(
                    symbol.domain_names
                )
            elif symbol.type == "negative":
                constraint += "\\leq 0 ~ \\forall " + ",".join(
                    symbol.domain_names
                )
            elif symbol.type == "sos1":
                constraint += "SOS1"
            elif symbol.type == "sos2":
                constraint += "SOS2"
            elif symbol.type == "semicont":
                constraint += "SemiCont"
            elif symbol.type == "semiint":
                constraint += r"\{0, 1, 2, ... \}" + " SemiInt"
            else:
                continue

            constraint += "\\\\$"
            constraints.append(constraint)

        return "\n".join(constraints)

    def get_header(self) -> str:
        header = """\\documentclass[11pt]{article}
\\usepackage{geometry}
\\usepackage[american]{babel}
\\usepackage{amsmath}
\\usepackage{amssymb}
\\usepackage[hidelinks]{hyperref}
\\usepackage{tabularx}
\\usepackage{ltablex}
\\keepXColumns

\\begin{document}
\\section*{Symbols}

"""
        return header



from __future__ import annotations

import os
import shutil
import tempfile

from gamspy.exceptions import ValidationError

DEBUGGING_LEVELS = ["delete", "keep_on_error", "keep"]


def validate_arguments(
    working_directory: str | None,
    debugging_level: str,
):
    # Validate working_directory
    if working_directory == "":
        raise ValidationError("`working_directory` cannot be an empty string.")

    # Validate debug_level
    if (
        not isinstance(debugging_level, str)
        or debugging_level not in DEBUGGING_LEVELS
    ):
        raise ValidationError(
            f"debugging level must be one of {DEBUGGING_LEVELS}"
        )


class Workspace:
    def __init__(
        self, debugging_level: str, working_directory: str | None = None
    ):
        validate_arguments(working_directory, debugging_level)

        self.debugging_level = debugging_level
        self.using_tmp_working_dir = False
        self._has_error = False

        if working_directory is None:
            self.using_tmp_working_dir = True
            self.working_directory = tempfile.mkdtemp()
        else:
            self.working_directory = os.path.abspath(working_directory)
            os.makedirs(self.working_directory, exist_ok=True)

    def __del__(self):
        if (
            hasattr(self, "using_tmp_working_dir")
            and self.using_tmp_working_dir
        ):
            try:
                if self.debugging_level == "delete":
                    shutil.rmtree(self.working_directory)

                if (
                    self.debugging_level == "keep_on_error"
                    and not self._has_error
                ):
                    shutil.rmtree(self.working_directory)
            except Exception:  # pragma: no cover
                ...



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy._algebra.expression import Expression


class ExtrinsicFunction(operable.Operable):
    """Extrinsic function registered by the user."""

    def __init__(self, name: str):
        self.name = name
        self.args: tuple = ()
        self.domain: list = []

    def __len__(self):
        return len(self.__str__())

    def __call__(self, *args, **kwds) -> Expression:
        if kwds:
            raise ValidationError(
                "External functions do not accept keyword arguments"
            )

        self.args = args
        return expression.Expression(None, self, None)

    def __str__(self) -> str:
        representation = self.name

        if self.args:
            arg_strs = []
            for arg in self.args:
                arg_str = (
                    str(arg)
                    if isinstance(arg, (int, float, str))
                    else arg.gamsRepr()
                )
                arg_strs.append(arg_str)

            args = ",".join(arg_strs)

            representation = f"{representation}({args})"

        return representation

    def gamsRepr(self) -> str:
        """
        Representation of this ExtrinsicFunction in GAMS language.

        Returns
        -------
        str
        """
        return self.__str__()

    def getDeclaration(self) -> str:
        """
        Declaration of the ExtrinsicFunction in GAMS

        Returns
        -------
        str
        """
        return self.gamsRepr()

    def latexRepr(self) -> str:
        """
        Representation of this ExtrinsicFunction in Latex.

        Returns
        -------
        str
        """
        return self.gamsRepr()


class ExtrinsicLibrary:
    """
    Extrinsic library registered by the user.

    Parameters
    ----------
    path : str
        Path to the shared object that contains the library.
    functions : dict[str, str]
        Functions to be imported from the extrinsic library.
    """

    def __init__(self, path: str, functions: dict[str, str]):
        self.name = "ext" + str(uuid.uuid4()).replace("-", "_")
        self.path = path
        self.functions = functions

    def getDeclaration(self) -> str:
        """
        Declaration of the ExtrinsicLibrary in GAMS

        Returns
        -------
        str
        """
        library_str = f"$funclibin {self.name} {self.path}\n"

        for gamspy_name, actual_name in self.functions.items():
            library_str += (
                f"function {gamspy_name} / {self.name}.{actual_name} /;\n"
            )

        return library_str

    def __getattr__(self, name):
        if name in self.functions:
            return ExtrinsicFunction(name)

        return self.__getattribute__(name)



from __future__ import annotations

import io
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, Literal

from gams.transfer._internals import GAMS_SYMBOL_MAX_LENGTH

import gamspy._symbols as symbols
import gamspy._symbols.implicits as implicits
import gamspy.utils as utils
from gamspy._model import Problem, Sense
from gamspy._options import EXECUTION_OPTIONS, MODEL_ATTR_OPTION_MAP, Options
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Equation, Model, Parameter, Set, Variable
    from gamspy._algebra.operation import Operation
    from gamspy._symbols.implicits import (
        ImplicitParameter,
        ImplicitSet,
        ImplicitVariable,
    )
    from gamspy._types import EllipsisType


def get_dimension(
    domain: Sequence[Set | Alias | ImplicitSet | str],
):
    dimension = 0

    for elem in domain:
        if isinstance(
            elem, (symbols.Set, symbols.Alias, implicits.ImplicitSet)
        ):
            dimension += elem.dimension
        else:
            dimension += 1

    return dimension


def get_domain_path(symbol: Set | Alias | ImplicitSet) -> list[str]:
    path: list[str] = []
    domain = symbol

    while domain != "*":
        if isinstance(domain, str):
            path.insert(0, domain)
        else:
            path.insert(0, domain.name)

        if isinstance(domain, symbols.Alias):
            path.insert(0, domain.alias_with.name)

        domain = "*" if isinstance(domain, str) else domain.domain[0]  # type: ignore

    return path


def validate_dimension(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    domain: list[Set | Alias | ImplicitSet | str],
):
    dimension = get_dimension(domain)

    entity_name = (
        f"symbol {symbol.name}"
        if hasattr(symbol, "name")
        else symbol.__class__.__name__
    )
    if dimension != symbol.dimension:
        raise ValidationError(
            f"The {entity_name} is referenced with"
            f" {'more' if dimension > symbol.dimension else 'less'} indices"
            f" than declared. Declared dimension is {symbol.dimension} but"
            f" given dimension is {dimension}"
        )


def validate_one_dimensional_sets(
    given: Set | Alias | ImplicitSet,
    actual: str | Set | Alias,
):
    if isinstance(given, implicits.ImplicitSet):
        return

    given_path = get_domain_path(given)

    if (
        isinstance(actual, symbols.Set)
        and actual.name not in given_path
        or (
            isinstance(actual, symbols.Alias)
            and actual.alias_with.name not in given_path
        )
    ):
        raise ValidationError(
            f"`Given set `{given.name}` is not a valid domain for declared"
            f" domain `{actual.name}`"
        )


def validate_type(domain):
    for given in domain:
        if not isinstance(
            given,
            (
                symbols.Set,
                symbols.Alias,
                symbols.UniverseAlias,
                implicits.ImplicitSet,
                str,
                type(...),
                slice,
            ),
        ):
            raise TypeError(
                "Domain item must be type Set, Alias, ImplicitSet or str but"
                f" found `{type(given)}`"
            )


def _get_ellipsis_range(domain, given_domain):
    start = 0
    end = len(domain)

    for item in given_domain:
        if isinstance(item, type(...)):
            break

        start += 1

    for item in reversed(given_domain):
        if isinstance(item, type(...)):
            break

        end -= 1

    return start, end


def _transform_given_indices(
    domain: list[Set | Alias | str],
    indices: EllipsisType | slice | Set | Alias | str | Iterable | ImplicitSet,
):
    new_domain: list = []
    given_domain = utils._to_list(indices)
    validate_type(given_domain)

    if len(domain) == 0:
        # If scalar, only correct indexing is [:] or [...]
        if len(given_domain) != 1:
            raise ValidationError(
                "Scalar values can only be indexed by '[:]' or '[...]'"
            )

        if not isinstance(given_domain[0], (type(...), slice)):
            raise ValidationError(
                "Scalar values can only be indexed by '[:]' or '[...]'"
            )

        return new_domain

    if len([item for item in given_domain if isinstance(item, type(...))]) > 1:
        raise ValidationError(
            "There cannot be more than one ellipsis in indexing"
        )

    index = 0
    for item in given_domain:
        dimension = (
            1 if isinstance(item, (str, type(...), slice)) else item.dimension
        )
        if isinstance(item, type(...)):
            start, end = _get_ellipsis_range(domain, given_domain)
            new_domain += domain[start:end]
            index = end
        elif isinstance(item, slice):
            new_domain.append(domain[index])
            index += dimension
        else:
            new_domain.append(item)
            index += dimension

    return new_domain


def validate_domain(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    indices: EllipsisType | slice | Set | Alias | str | Iterable | ImplicitSet,
):
    domain = _transform_given_indices(symbol.domain, indices)  # type: ignore
    validate_container(symbol, domain)
    validate_dimension(symbol, domain)

    offset = 0
    for given in domain:
        given_dim = 1 if isinstance(given, str) else given.dimension
        actual = symbol.domain[offset]
        actual_dim = 1 if isinstance(actual, str) else actual.dimension

        if actual_dim == 1 and given_dim == 1:
            if isinstance(given, str):
                if (
                    hasattr(actual, "records")
                    and len(actual.records) < 1000
                    and not actual.records.isin([given]).sum().any()
                ):
                    raise ValidationError(
                        f"Literal index `{given}` was not found in set"
                        f" `{actual.name}`"
                    )
            else:
                validate_one_dimensional_sets(given, actual)

        offset += given_dim

    return domain


def validate_container(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    domain: list[str | Set | Alias],
):
    for set in domain:
        if (
            isinstance(set, (symbols.Set, symbols.Alias))
            and set.container != symbol.container
        ):
            raise ValidationError(
                f"`Domain `{set}` must be in the same container"
                f" with `{symbol}`"
            )


def validate_name(word: str) -> str:
    if not isinstance(word, str):
        raise TypeError("Symbol name must be type str")

    reserved_words = [
        "abort",
        "acronym",
        "acronyms",
        "alias",
        "all",
        "and",
        "binary",
        "break",
        "card",
        "continue",
        "diag",
        "display",
        "do",
        "else",
        "elseif",
        "endfor",
        "endif",
        "endloop",
        "endwhile",
        "eps",
        "equation",
        "equations",
        "execute",
        "execute_load",
        "execute_loaddc",
        "execute_loadhandle",
        "execute_loadpoint",
        "execute_unload",
        "execute_unloaddi",
        "execute_unloadidx",
        "file",
        "files",
        "for",
        "free",
        "function",
        "functions",
        "gdxLoad",
        "if",
        "inf",
        "integer",
        "logic",
        "loop",
        "model",
        "models",
        "na",
        "negative",
        "nonnegative",
        "no",
        "not",
        "option",
        "options",
        "or",
        "ord",
        "parameter",
        "parameters",
        "positive",
        "prod",
        "put",
        "put_utility",
        "putclear",
        "putclose",
        "putfmcl",
        "puthd",
        "putheader",
        "putpage",
        "puttitle",
        "puttl",
        "repeat",
        "sameas",
        "sand",
        "scalar",
        "scalars",
        "semicont",
        "semiint",
        "set",
        "sets",
        "singleton",
        "smax",
        "smin",
        "solve",
        "sor",
        "sos1",
        "sos2",
        "sum",
        "system",
        "table",
        "tables",
        "then",
        "undf",
        "until",
        "variable",
        "variables",
        "while",
        "xor",
        "yes",
    ]

    if word.lower() in reserved_words:
        raise ValidationError(
            "Name cannot be one of the reserved words. List of reserved"
            f" words: {reserved_words}"
        )

    return word


def validate_model(
    equations: Iterable[Equation],
    problem: Problem | str,
    sense: str | Sense,
) -> tuple[Problem, Sense]:
    if isinstance(problem, str):
        if problem.upper() not in Problem.values():
            raise ValueError(
                f"Allowed problem types: {Problem.values()} but found"
                f" {problem}."
            )

        problem = Problem(problem.upper())

    if isinstance(sense, str):
        if sense.upper() not in Sense.values():
            raise ValueError(
                f"Allowed sense values: {Sense.values()} but found"
                f" {sense}."
            )

        sense = Sense(sense.upper())

    if (
        problem not in (Problem.CNS, Problem.MCP)
        and not isinstance(equations, Iterable)
        or any(
            not isinstance(equation, symbols.Equation)
            for equation in equations
        )
    ):
        raise TypeError("`equations` must be an Iterable of Equation objects")

    return problem, sense


def validate_model_name(name: str) -> str:
    if len(name) == 0:
        raise ValueError("Model name must be at least one character.")

    if len(name) > GAMS_SYMBOL_MAX_LENGTH:
        raise ValueError(
            "Model 'name' is too long, "
            f"max is {GAMS_SYMBOL_MAX_LENGTH} characters"
        )

    if name[0] == "_":
        raise ValidationError(
            "Valid GAMS names cannot begin with a '_' character."
        )

    if not all(True if i == "_" else i.isalnum() for i in name):
        raise ValidationError(
            f"`{name}` is an invalid model name. "
            "Model names can only contain alphanumeric characters "
            "(letters and numbers) and the '_' character."
        )

    return name


def validate_solver_args(
    system_directory: str,
    backend: Literal["local", "engine", "neos"],
    solver: str,
    problem: Problem | str,
    options: Options | None,
    output: io.TextIOWrapper | None,
    load_symbols: list[Symbol] | None,
) -> None:
    # Check validity of options
    if options is not None and not isinstance(options, Options):
        raise TypeError(
            f"`options` must be of type Option but found {type(options)}"
        )

    # Check validity of output
    if output is not None and (
        not hasattr(output, "write") or not hasattr(output, "flush")
    ):
        raise ValidationError(
            "`output` must write and flush operations but found"
            f" {type(output)} which does not support them."
        )

    # Check validity of load_symbols
    if load_symbols is not None:
        if not isinstance(load_symbols, list):
            raise ValidationError(
                f"`load_symbols` must be list of Symbol objects. Given type: {type(load_symbols)}"
            )

        for elem in load_symbols:
            if not isinstance(elem, Symbol):
                raise ValidationError(
                    f"Elements of `load_symbols` must be of type Symbol but found {elem}"
                )

    # Check validity of solver
    if not isinstance(solver, str):
        raise TypeError("`solver` argument must be a string.")

    if backend == "neos" and solver.lower() in ["mpsge", "kestrel"]:
        raise ValidationError(
            f"`{solver}` is not a valid solver for NEOS Server."
        )

    if backend == "engine" and solver.lower() in ["mpsge"]:
        raise ValidationError(
            f"`{solver}` is not a valid solver for GAMS Engine."
        )

    if backend == "local":
        # No need to check whether the solver is installed on client's machine for NEOS or ENGINE.
        installed_solvers = utils.getInstalledSolvers(system_directory)
        if solver.upper() not in installed_solvers:
            raise ValidationError(
                f"Provided solver name `{solver}` is not installed on your"
                f" machine. Install `{solver}` with `gamspy install solver"
                f" {solver.lower()}`"
            )

        capabilities = utils.getSolverCapabilities(system_directory)
        if str(problem).upper() not in capabilities[solver.upper()]:
            raise ValidationError(
                f"Given solver `{solver}` is not capable of solving given"
                f" problem type `{problem}`. See capability matrix "
                "(https://www.gams.com/latest/docs/S_MAIN.html#SOLVERS_MODEL_TYPES)"
                " to choose a suitable solver"
            )


def validate_equations(model: Model):
    for equation in model.equations:
        if equation._definition is None:
            raise ValidationError(
                f"`{equation.name}` has been declared as an equation but no equation definition was found."
            )


def validate_global_options(options: Any) -> Options:
    if options is not None and not isinstance(options, Options):
        raise TypeError(
            f"`options` must be of type Option but found {type(options)}"
        )

    if isinstance(options, Options):
        options_dict = options.model_dump(exclude_none=True)
        if any(option in options_dict for option in MODEL_ATTR_OPTION_MAP):
            raise ValidationError(
                f"The following model options cannot be provided at Container creation time: {', '.join(MODEL_ATTR_OPTION_MAP.keys())}."
            )

        if any(option in options_dict for option in EXECUTION_OPTIONS):
            raise ValidationError(
                f"The following runtime options cannot be provided at Container creation time: {', '.join(EXECUTION_OPTIONS.keys())}."
            )

    if options is None:
        return Options()

    return options



from __future__ import annotations

from abc import ABC, abstractmethod

import gams.transfer as gt

import gamspy._algebra.condition as condition
import gamspy.utils as utils
from gamspy.exceptions import ValidationError


class ImplicitSymbol(ABC):
    def __init__(
        self,
        parent,
        name,
        domain,
        permutation=None,
        parent_scalar_domains=None,
    ) -> None:
        self.parent = parent
        self.container = parent.container
        self.name = name
        self.domain = domain
        self.where = condition.Condition(self)

        if permutation is not None:
            permutation = list(permutation)

        self.permutation = permutation

        if parent_scalar_domains is None:
            parent_scalar_domains = []
        else:
            parent_scalar_domains = list(parent_scalar_domains)

        self.parent_scalar_domains = parent_scalar_domains

        self.fix_domain_scalars(parent_scalar_domains)

    def __bool__(self):
        raise ValidationError("A symbol cannot be used as a truth value.")

    def fix_domain_scalars(self, parent_scalar_domains):
        if len(self.domain) == 1 and self.domain[0] == "*":
            self._scalar_domains = []
            return

        bare_domain = utils._get_set(self.domain)
        domain = []
        scalars = []
        permutation_indices_to_del = []
        for i, d in enumerate(bare_domain):
            if isinstance(d, str):
                loc_scalar = (
                    i if self.permutation is None else self.permutation[i]
                )
                scalars.append((loc_scalar, d))
                permutation_indices_to_del.append(i)
            else:
                domain.append(d)

        permutation_indices_to_del = reversed(permutation_indices_to_del)
        if self.permutation is not None:
            for index in permutation_indices_to_del:
                del self.permutation[index]

            self.fix_permutation()

        self.domain = domain

        scalars.extend(parent_scalar_domains)
        scalars = list(sorted(scalars, key=lambda k: k[0]))
        self._scalar_domains = scalars

    def fix_permutation(self):
        numbers_pos = list(zip(self.permutation, range(len(self.permutation))))
        numbers_pos = sorted(numbers_pos, key=lambda k: k[0])
        new_perm = [-1] * len(self.permutation)
        for i, (_, pos) in enumerate(numbers_pos):
            new_perm[pos] = i

        self.permutation = new_perm

    @property
    def dimension(self) -> int:
        return self.parent.dimension - len(self._scalar_domains)

    @abstractmethod
    def gamsRepr(self):
        """Representation of the implicit symbol in GAMS"""

    def latexRepr(self):
        from .implicit_set import ImplicitSet

        name = self.name.replace("_", "\\_")
        representation = f"{name}"
        domain = list(self.domain)

        if hasattr(self, "_scalar_domains"):
            for i, d in self._scalar_domains:
                domain.insert(i, d)

        if domain:
            set_strs = []
            for elem in domain:
                if isinstance(elem, (gt.Set, gt.Alias, ImplicitSet)):
                    set_strs.append(elem.latexRepr())
                elif isinstance(elem, str):
                    set_strs.append("*")

            domain_str = ",".join(set_strs)
            representation = f"{representation}_{{{domain_str}}}"

        return representation



from __future__ import annotations

import math

import gamspy as gp
from gamspy.exceptions import ValidationError


def _check_tuple_int(
    value: int | tuple[int, int], name: str, allow_zero=False
) -> tuple[int, int]:
    if not isinstance(value, (int, tuple)):
        raise ValidationError(
            f"{name} must be an integer or a tuple of integer"
        )

    if isinstance(value, int):
        value = (value, value)

    cmp = (lambda a: a >= 0) if allow_zero else (lambda a: a > 0)
    text = "or equal to " if allow_zero else ""

    if not (isinstance(value[0], int) and cmp(value[0])):
        raise ValidationError(f"{name} must be greater than {text}0")

    if not (isinstance(value[1], int) and cmp(value[1])):
        raise ValidationError(f"{name} must be a greater than {text}0")

    return value


def _calc_same_padding(
    kernel_size: tuple[int, int],
    h_in: int,
    w_in: int,
) -> tuple[int, int]:
    # assumes stride = 1
    pad_h = math.floor((kernel_size[0] - 1) / 2)
    pad_w = math.floor((kernel_size[1] - 1) / 2)
    return pad_h, pad_w


def _calc_hw(
    padding: tuple[int, int] | str,
    kernel_size: tuple[int, int],
    stride: tuple[int, int],
    h_in: int,
    w_in: int,
) -> tuple[int, int]:
    if isinstance(padding, str):
        return h_in, w_in

    h_out = math.floor(
        1 + ((h_in + (2 * padding[0]) - (kernel_size[0] - 1) - 1) / stride[0])
    )
    w_out = math.floor(
        1 + ((w_in + (2 * padding[1]) - (kernel_size[1] - 1) - 1) / stride[1])
    )
    return h_out, w_out


def _next_domains(
    input_domain: list[gp.Set], check_domains: list[gp.Set]
) -> list[gp.Set]:
    names = {x.name for x in check_domains}
    output = []
    for domain in input_domain:
        while domain.name in names:
            domain = gp.math.next_alias(domain)

        output.append(domain)
        names.add(domain.name)

    return output



from __future__ import annotations

import uuid

import gamspy as gp
import gamspy.formulations.nn.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim


class AvgPool2d:
    """
    Formulation generator for 2D Avg Pooling in GAMS.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    kernel_size : int | tuple[int, int]
        Filter size
    stride : int | tuple[int, int] | None
        Stride in the avg pooling, it is equal to kernel_size if not provided
    padding : int | tuple[int, int]
        Amount of padding to be added to input, by default 0

    Examples
    --------
    >>> import gamspy as gp
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> # 2x2 avg pooling
    >>> ap1 = gp.formulations.AvgPool2d(m, (2, 2))
    >>> inp = gp.Variable(m, domain=dim((10, 1, 24, 24)))
    >>> out, eqs = ap1(inp)
    >>> type(out)
    <class 'gamspy._symbols.variable.Variable'>
    >>> [len(x) for x in out.domain]
    [10, 1, 12, 12]

    """

    def __init__(
        self,
        container: gp.Container,
        kernel_size: int | tuple[int, int],
        stride: int | tuple[int, int] | None = None,
        padding: int = 0,
    ):
        _kernel_size = utils._check_tuple_int(kernel_size, "kernel_size")
        if stride is None:
            stride = kernel_size

        _stride = utils._check_tuple_int(stride, "stride")
        _padding = utils._check_tuple_int(padding, "padding", allow_zero=True)

        self.container = container
        self.kernel_size = _kernel_size
        self.stride = _stride
        self.padding = _padding

    def _get_bounds(
        self, input: gp.Parameter | gp.Variable
    ) -> tuple[gp.Parameter, gp.Parameter]:
        # this can be done more fine-grained
        N, C, H, W = input.domain
        lb = gp.Parameter(self.container, domain=[N, C])
        ub = gp.Parameter(self.container, domain=[N, C])

        if isinstance(input, gp.Variable):
            ub[...] = gp.Smax([H, W], input.up[N, C, H, W])
            lb[...] = gp.Smin([H, W], input.lo[N, C, H, W])
        else:
            ub[...] = gp.Smax([H, W], input[N, C, H, W])
            lb[...] = gp.Smin([H, W], input[N, C, H, W])

        # 0 padding on the edges can shift the minimum and max value
        scale = (
            (self.kernel_size[0] - self.padding[0])
            * (self.kernel_size[1] - self.padding[1])
        ) / (self.kernel_size[0] * self.kernel_size[1])

        lb[N, C].where[lb[N, C] > 0] = lb[N, C] * scale
        ub[N, C].where[ub[N, C] < 0] = ub[N, C] * scale

        return lb, ub

    def __call__(
        self, input: gp.Parameter | gp.Variable
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the average pooling. Unlike the min or max pooling avg
        pooling does not require binary variables or the big-M formulation.
        Returns the output variable and the list of equations required for
        the avg pooling formulation

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the max pooling 2d layer, must be in shape
                (batch x in_channels x height x width)

        Returns
        -------
        tuple[gp.Variable, list[gp.Equation]]

        """
        if not isinstance(input, (gp.Parameter, gp.Variable)):
            raise ValidationError("Expected a parameter or a variable input")

        if len(input.domain) != 4:
            raise ValidationError(
                f"expected 4D input (got {len(input.domain)}D input)"
            )

        N, C_in, H_in, W_in = input.domain

        h_in = len(H_in)
        w_in = len(W_in)

        h_out, w_out = utils._calc_hw(
            self.padding, self.kernel_size, self.stride, h_in, w_in
        )

        lb, ub = self._get_bounds(input)
        out_var = gp.Variable(
            self.container, domain=dim([len(N), len(C_in), h_out, w_out])
        )
        out_var.lo[...] = lb
        out_var.up[...] = ub

        N, C, H_out, W_out = out_var.domain

        set_out = gp.Equation(self.container, domain=out_var.domain)

        # expr must have domain N, C, H_out, W_out
        top_index = (
            (self.stride[0] * (gp.Ord(H_out) - 1)) - self.padding[0] + 1
        )
        left_index = (
            (self.stride[1] * (gp.Ord(W_out) - 1)) - self.padding[1] + 1
        )
        coeff = 1 / (self.kernel_size[0] * self.kernel_size[1])

        Hf, Wf = gp.math._generate_dims(self.container, self.kernel_size)
        Hf, Wf, H_in, W_in = utils._next_domains(
            [Hf, Wf, H_in, W_in], out_var.domain
        )

        name = "ds_" + str(uuid.uuid4()).split("-")[0]
        subset = gp.Set(
            self.container, name, domain=[H_out, W_out, Hf, Wf, H_in, W_in]
        )
        subset[
            H_out,
            W_out,
            Hf,
            Wf,
            H_in,
            W_in,
        ].where[
            (gp.Ord(H_in) == (top_index + gp.Ord(Hf) - 1))
            & (gp.Ord(W_in) == (left_index + gp.Ord(Wf) - 1))
        ] = True

        expr = gp.Sum(
            subset[H_out, W_out, Hf, Wf, H_in, W_in],
            input[N, C, H_in, W_in] * coeff,
        )

        set_out[...] = out_var[N, C, H_out, W_out] == expr
        return out_var, [set_out]



from __future__ import annotations

import doctest
import glob
import json
import os
import shutil
from os.path import join

import pytest

import gamspy


@pytest.mark.unit
def test_version():
    assert gamspy.__version__ == "1.5.1"


@pytest.mark.doc
def test_switcher():
    this = os.path.dirname(os.path.abspath(__file__))
    root = this.rsplit(os.sep, maxsplit=1)[0]
    with open(join(root, "docs", "_static", "switcher.json")) as file:
        switcher = json.loads(file.read())
        versions = [elem["version"] for elem in switcher]
        assert f"v{gamspy.__version__}" in versions


@pytest.fixture
def teardown():
    # Act and assert
    yield

    # Cleanup
    files = glob.glob("_*")
    for file in files:
        os.remove(file)

    if os.path.exists("test"):
        shutil.rmtree("test")


@pytest.mark.doc
def test_docs():
    api_files = [
        file
        for file in glob.glob("src/**", recursive=True)
        if file.endswith(".py")
    ]

    for file in api_files:
        results = doctest.testfile(
            file,
            verbose=True,
            module_relative=False,
        )

        assert results.failed == 0



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union, no_type_check

import gams.transfer as gt

import gamspy._symbols.implicits as implicits
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Equation, Parameter, Set, Variable

    SymbolType = Union[Alias, Set, Parameter, Variable, Equation]


class Symbol(ABC):
    def __bool__(self):
        raise ValidationError("A symbol cannot be used as a truth value.")

    @abstractmethod
    def gamsRepr(self):
        """Representation of the implicit symbol in GAMS"""

    def latexRepr(self):
        """
        Representation of symbol in Latex.

        Returns
        -------
        str
        """
        name = self.name.replace("_", "\\_")
        representation = f"{name}"
        domain = list(self.domain)

        if hasattr(self, "_scalar_domains"):
            for i, d in self._scalar_domains:
                domain.insert(i, d)

        if domain and domain != ["*"]:
            set_strs = []
            for elem in domain:
                if isinstance(elem, (gt.Set, gt.Alias, implicits.ImplicitSet)):
                    set_strs.append(elem.latexRepr())
                elif isinstance(elem, str):
                    set_strs.append("*")

            domain_str = ",".join(set_strs)
            representation = f"{representation}_{{{domain_str}}}"

        return representation

    @property
    def synchronize(self) -> bool:
        """
        Synchronization state of the symbol. If True, the symbol data
        will be communicated with GAMS. Otherwise, GAMS state will not be updated.

        Returns
        -------
        bool

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=["i1"])
        >>> i.synchronize = False
        >>> i["i2"] = True
        >>> i.records.uni.tolist()
        ['i1']
        >>> i.synchronize = True
        >>> i.records.uni.tolist()
        ['i1', 'i2']

        """
        return self._synchronize

    @synchronize.setter
    @no_type_check
    def synchronize(self, value: bool):
        """
        If set to True, the current state will be synchronized with GAMS.
        Else, the symbol will not be synchronized with GAMS.
        """
        if value:
            self._synchronize = True

            if self._winner == "python":
                self.modified = True
                self.container._synch_with_gams()
            else:
                self.container.loadRecordsFromGdx(
                    load_from=self.container._gdx_out, symbol_names=[self.name]
                )
        else:
            self._synchronize = False

    @no_type_check
    def _get_domain_str(self, forwardings: bool | list[bool]):
        if isinstance(forwardings, bool):
            forwardings = [forwardings] * self.dimension

        set_strs = []
        for elem, forwarding in zip(self.domain, forwardings):
            if isinstance(elem, (gt.Set, gt.Alias, implicits.ImplicitSet)):
                elem_str = elem.gamsRepr()
                if forwarding:
                    elem_str += "<"
                set_strs.append(elem_str)
            elif isinstance(elem, (str, gt.UniverseAlias)):
                set_strs.append("*")

        return "(" + ",".join(set_strs) + ")"
